import pytest
import requests

from cads_api_client import ApiClient, processing


@pytest.fixture
def proc(api_root_url: str, api_anon_key: str) -> processing.Processing:
    client = ApiClient(url=api_root_url, key=api_anon_key, maximum_tries=0)
    return client.retrieve_api


def test_processes(proc: processing.Processing) -> None:
    res = proc.processes()

    assert isinstance(res, processing.ProcessList)
    assert "processes" in res.json
    assert isinstance(res.json["processes"], list)
    assert "links" in res.json
    assert isinstance(res.json["links"], list)

    assert len(res.process_ids()) == 10


def test_processes_limit(proc: processing.Processing) -> None:
    processes = proc.processes(params={"limit": 1})

    res = processes.next()

    if res is not None:
        assert res.response.status_code == 200


def test_process(proc: processing.Processing) -> None:
    process_id = "test-adaptor-dummy"

    res = proc.process(process_id)

    assert isinstance(res, processing.Process)
    assert res.id == process_id
    assert "links" in res.json
    assert isinstance(res.json["links"], list)


def test_validate_constraints(proc: processing.Processing) -> None:
    process_id = "test-adaptor-mars"
    process = proc.process(process_id)
    res = process.valid_values({})

    assert set(["product_type", "variable", "year", "month", "time"]) <= set(res)


def test_collection_anonymous_user(proc: processing.Processing) -> None:
    collection_id = "test-adaptor-mars"
    process = proc.process(collection_id)
    response = process.execute(inputs={})
    assert "message" in response.json


def test_jobs_list(proc: processing.Processing) -> None:
    collection_id = "test-adaptor-dummy"
    process = proc.process(collection_id)

    _ = process.execute(inputs={})
    _ = process.execute(inputs={})

    res = proc.jobs().json
    assert len(res["jobs"]) >= 2

    res = proc.jobs(params={"limit": 1}).json
    assert len(res["jobs"]) == 1

    jobs = proc.jobs(params={"limit": 1})
    res = jobs.next().json  # type: ignore

    assert res is not None
    assert len(res["jobs"]) == 1


def test_validate_constraints_error(proc: processing.Processing) -> None:
    process_id = "test-adaptor-mars"
    process = proc.process(process_id)
    with pytest.raises(requests.exceptions.HTTPError, match="422 Client Error") as exc:
        process.valid_values({"invalid_param": 1})
        assert exc.response.status_code == 422  # type: ignore[attr-defined]



import json
import pathlib

import pytest

from cads_api_client import config


def test_read_configuration(
    tmp_path: pathlib.Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    expected_config = {"url": "dummy-url", "key": "dummy-key"}

    config_file = tmp_path / ".cads-api-client.json"
    with config_file.open("w") as fp:
        json.dump({"url": "dummy-url", "key": "dummy-key"}, fp)

    res = config.read_configuration_file(str(config_file))
    assert res == expected_config

    monkeypatch.setenv("CADS_API_RC", str(config_file))
    res = config.read_configuration_file(None)
    assert res == expected_config


def test_read_configuration_error(tmp_path: pathlib.Path) -> None:
    config_file = tmp_path / ".cads-api-client.json"
    config_file.write_text("XXX")

    with pytest.raises(ValueError):
        config.read_configuration_file(str(config_file))

    with pytest.raises(FileNotFoundError):
        config.read_configuration_file("non-existent-file")


def test_get_config_from_configuration_file(tmp_path: pathlib.Path) -> None:
    expected_config = {"url": "dummy-url", "key": "dummy-key"}

    config_file = tmp_path / ".cads-api-client.json"
    with config_file.open("w") as fp:
        json.dump(expected_config, fp)

    res = config.get_config("url", str(config_file))

    assert res == expected_config["url"]

    with pytest.raises(KeyError):
        config.get_config("non-existent-key", str(config_file))


def test_get_config_from_environment_variables(
    tmp_path: pathlib.Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    expected_config = {"url": "dummy-url", "key": "dummy-key"}
    file_config = {"url": "wrong-url", "key": "wrong-key"}

    config_file = tmp_path / ".cads-api-client.json"
    with config_file.open("w") as fp:
        json.dump(file_config, fp)

    monkeypatch.setenv("CADS_API_URL", expected_config["url"])
    monkeypatch.setenv("CADS_API_KEY", expected_config["key"])

    res = config.get_config("url", str(config_file))

    assert res == expected_config["url"]

    res = config.get_config("key", str(config_file))

    assert res == expected_config["key"]



import pytest

from cads_api_client import ApiClient, catalogue


@pytest.fixture
def cat(api_root_url: str, monkeypatch: pytest.MonkeyPatch) -> catalogue.Catalogue:
    monkeypatch.setenv("CADS_API_RC", "")
    monkeypatch.delenv("CADS_API_KEY", raising=False)
    with pytest.warns(UserWarning, match="The API key is missing"):
        client = ApiClient(url=api_root_url, maximum_tries=0)
    return client.catalogue_api


def test_collections(cat: catalogue.Catalogue) -> None:
    res: catalogue.Collections | None = cat.collections()

    assert isinstance(res, catalogue.Collections)
    assert "collections" in res.json
    assert isinstance(res.json["collections"], list)
    assert "links" in res.json
    assert isinstance(res.json["links"], list)

    collection_ids = res.collection_ids()
    while len(collection_ids) != res.json["numberMatched"]:
        res = res.next()
        assert res is not None
        collection_ids.extend(res.collection_ids())

    expected_collection_id = "reanalysis-era5-single-levels"
    assert expected_collection_id in collection_ids


def test_collections_limit(cat: catalogue.Catalogue) -> None:
    collections = cat.collections(params={"limit": 1})

    res = collections.next()

    if res is not None:
        assert res.response.status_code == 200


def test_collection(cat: catalogue.Catalogue) -> None:
    collection_id = "test-adaptor-mars"

    res = cat.collection(collection_id)

    assert isinstance(res, catalogue.Collection)
    assert "id" in res.json
    assert res.id == collection_id
    assert "links" in res.json
    assert isinstance(res.json["links"], list)
    assert res.begin_datetime.isoformat() == "1959-01-01T00:00:00+00:00"
    assert res.end_datetime.isoformat() == "2023-05-09T00:00:00+00:00"



from __future__ import annotations

import os

import pytest


@pytest.fixture
def api_root_url() -> str:
    from cads_api_client import config

    try:
        return str(config.get_config("url"))
    except Exception:
        return "http://localhost:8080/api"


@pytest.fixture
def api_anon_key() -> str:
    return os.getenv("CADS_API_ANON_KEY", "00112233-4455-6677-c899-aabbccddeeff")



from __future__ import annotations

import contextlib
import pathlib
import time
from typing import Any

import pytest
import requests

from cads_api_client import legacy_api_client, processing

does_not_raise = contextlib.nullcontext


def legacy_update(remote: processing.Remote) -> None:
    # See https://github.com/ecmwf/cdsapi/blob/master/examples/example-era5-update.py
    sleep = 1
    while True:
        with pytest.deprecated_call():
            remote.update()

        reply = remote.reply
        remote.info("Request ID: %s, state: %s" % (reply["request_id"], reply["state"]))

        if reply["state"] == "completed":
            break
        elif reply["state"] in ("queued", "running"):
            remote.info("Request ID: %s, sleep: %s", reply["request_id"], sleep)
            time.sleep(sleep)
        elif reply["state"] in ("failed",):
            remote.error("Message: %s", reply["error"].get("message"))
            remote.error("Reason:  %s", reply["error"].get("reason"))
            for n in (
                reply.get("error", {})
                .get("context", {})
                .get("traceback", "")
                .split("\n")
            ):
                if n.strip() == "":
                    break
                remote.error("  %s", n)
            raise Exception(
                "%s. %s."
                % (reply["error"].get("message"), reply["error"].get("reason"))
            )


def test_retrieve(tmp_path: pathlib.Path, api_root_url: str, api_anon_key: str) -> None:
    client = legacy_api_client.LegacyApiClient(
        url=api_root_url, key=api_anon_key, retry_max=0
    )

    collection_id = "test-adaptor-dummy"
    request = {"size": 1}

    target = tmp_path / "test-retrieve-with-target.grib"
    actual_target = client.retrieve(collection_id, request, str(target))
    assert str(target) == actual_target
    assert target.stat().st_size == 1

    result = client.retrieve(collection_id, request)
    target = tmp_path / "test-retrieve-no-target.grib"
    actual_target = result.download(str(target))
    assert str(target) == actual_target
    assert target.stat().st_size == 1

    response = requests.head(result.location)
    assert response.status_code == 200

    assert result.content_length == 1
    assert result.content_type == "application/x-grib"


@pytest.mark.parametrize("quiet", [True, False])
def test_quiet(
    caplog: pytest.LogCaptureFixture,
    api_root_url: str,
    api_anon_key: str,
    quiet: bool,
) -> None:
    client = legacy_api_client.LegacyApiClient(
        url=api_root_url, key=api_anon_key, quiet=quiet, retry_max=0
    )
    client.retrieve("test-adaptor-dummy", {})
    records = [record for record in caplog.records if record.levelname == "INFO"]
    assert not records if quiet else records


@pytest.mark.parametrize("debug", [True, False])
def test_debug(
    caplog: pytest.LogCaptureFixture,
    api_root_url: str,
    api_anon_key: str,
    debug: bool,
) -> None:
    legacy_api_client.LegacyApiClient(
        url=api_root_url, key=api_anon_key, debug=debug, retry_max=0
    )
    records = [record for record in caplog.records if record.levelname == "DEBUG"]
    assert records if debug else not records


@pytest.mark.parametrize(
    "wait_until_complete,expected_type",
    [(True, processing.Results), (False, processing.Remote)],
)
def test_wait_until_complete(
    tmp_path: pathlib.Path,
    api_root_url: str,
    api_anon_key: str,
    wait_until_complete: bool,
    expected_type: type,
) -> None:
    client = legacy_api_client.LegacyApiClient(
        url=api_root_url,
        key=api_anon_key,
        wait_until_complete=wait_until_complete,
        retry_max=0,
    )

    collection_id = "test-adaptor-dummy"
    request = {"size": 1}

    result = client.retrieve(collection_id, request)
    assert isinstance(result, expected_type)

    target = tmp_path / "test.grib"
    result.download(str(target))
    assert target.stat().st_size == 1


@pytest.mark.parametrize(
    "collection_id,raises",
    [
        ("test-adaptor-dummy", does_not_raise()),
        ("test-adaptor-mars", pytest.raises(Exception, match="400 Client Error")),
    ],
)
def test_legacy_update(
    api_root_url: str,
    api_anon_key: str,
    collection_id: str,
    raises: contextlib.nullcontext[Any],
) -> None:
    client = legacy_api_client.LegacyApiClient(
        url=api_root_url, key=api_anon_key, wait_until_complete=False, retry_max=0
    )
    remote = client.retrieve(collection_id, {})
    assert isinstance(remote, processing.Remote)
    with raises:
        legacy_update(remote)


def test_legacy_api_client_kwargs(api_root_url: str, api_anon_key: str) -> None:
    session = requests.Session()
    client = legacy_api_client.LegacyApiClient(
        url=api_root_url,
        key=api_anon_key,
        verify=False,
        timeout=1,
        progress=False,
        delete=True,
        retry_max=2,
        sleep_max=3,
        wait_until_complete=False,
        session=session,
    )
    assert client.client.url == api_root_url
    assert client.client.key == api_anon_key
    assert client.client.verify is False
    assert client.timeout == 1
    assert client.client.progress is False
    assert client.client.cleanup is True
    assert client.client.maximum_tries == 2
    assert client.client.sleep_max == 3


def test_legacy_api_client_error(
    api_root_url: str,
    api_anon_key: str,
) -> None:
    with pytest.raises(ValueError, match="Wrong parameters: {'foo'}"):
        legacy_api_client.LegacyApiClient(url=api_root_url, key=api_anon_key, foo="bar")



from __future__ import annotations

from typing import Any

import attrs
import requests

from . import config, processing


@attrs.define(slots=False)
class Profile:
    url: str
    headers: dict[str, Any]
    session: requests.Session
    retry_options: dict[str, Any]
    request_options: dict[str, Any]
    download_options: dict[str, Any]
    sleep_max: int
    cleanup: bool
    force_exact_url: bool = False

    def __attrs_post_init__(self) -> None:
        if not self.force_exact_url:
            self.url += f"/{config.SUPPORTED_API_VERSION}"

    @property
    def request_kwargs(self) -> processing.RequestKwargs:
        return processing.RequestKwargs(
            headers=self.headers,
            session=self.session,
            retry_options=self.retry_options,
            request_options=self.request_options,
            download_options=self.download_options,
            sleep_max=self.sleep_max,
            cleanup=self.cleanup,
        )

    def get_api_response(
        self, method: str, url: str, **kwargs: Any
    ) -> processing.ApiResponse:
        return processing.ApiResponse.from_request(
            method,
            url,
            **self.request_kwargs,
            **kwargs,
        )

    def profile(self) -> dict[str, Any]:
        url = f"{self.url}/account"
        return self.get_api_response("get", url).json

    def accept_licence(self, licence_id: str, revision: int) -> dict[str, Any]:
        url = f"{self.url}/account/licences/{licence_id}"
        return self.get_api_response("put", url, json={"revision": revision}).json

    def accepted_licences(self) -> dict[str, Any]:
        url = f"{self.url}/account/licences"
        return self.get_api_response("get", url).json

    def check_authentication(self) -> dict[str, Any]:
        url = f"{self.url}/account/verification/pat"
        return self.get_api_response("post", url).json



from __future__ import annotations

import functools
import warnings
from typing import Any

import attrs
import multiurl.base
import requests

from . import catalogue, config, processing, profile


def strtobool(value: str) -> bool:
    if value.lower() in ("y", "yes", "t", "true", "on", "1"):
        return True
    if value.lower() in ("n", "no", "f", "false", "off", "0"):
        return False
    raise ValueError(f"invalid truth value {value!r}")


@attrs.define(slots=False)
class ApiClient:
    url: str | None = None
    key: str | None = None
    verify: bool | None = None
    timeout: int = 60
    progress: bool = True
    cleanup: bool = False
    sleep_max: int = 120
    retry_after: int = 120
    maximum_tries: int = 500
    session: requests.Session = attrs.field(factory=requests.Session)

    def __attrs_post_init__(self) -> None:
        if self.url is None:
            self.url = str(config.get_config("url"))

        if self.key is None:
            try:
                self.key = str(config.get_config("key"))
            except (KeyError, FileNotFoundError):
                warnings.warn("The API key is missing", UserWarning)

        if self.verify is None:
            try:
                self.verify = strtobool(str(config.get_config("verify")))
            except (KeyError, FileNotFoundError):
                self.verify = True

    def _get_headers(self, key_is_mandatory: bool = True) -> dict[str, str]:
        if self.key is None:
            if key_is_mandatory:
                raise ValueError("The API key is needed to access this resource")
            return {}
        return {"PRIVATE-TOKEN": self.key}

    @property
    def _retry_options(self) -> dict[str, Any]:
        return {
            "maximum_tries": self.maximum_tries,
            "retry_after": self.retry_after,
        }

    @property
    def _download_options(self) -> dict[str, Any]:
        progress_bar = (
            multiurl.base.progress_bar if self.progress else multiurl.base.NoBar
        )
        return {
            "progress_bar": progress_bar,
        }

    @property
    def _request_options(self) -> dict[str, Any]:
        return {
            "timeout": self.timeout,
            "verify": self.verify,
        }

    def _get_request_kwargs(
        self, mandatory_key: bool = True
    ) -> processing.RequestKwargs:
        return processing.RequestKwargs(
            headers=self._get_headers(key_is_mandatory=mandatory_key),
            session=self.session,
            retry_options=self._retry_options,
            request_options=self._request_options,
            download_options=self._download_options,
            sleep_max=self.sleep_max,
            cleanup=self.cleanup,
        )

    @functools.cached_property
    def catalogue_api(self) -> catalogue.Catalogue:
        return catalogue.Catalogue(
            f"{self.url}/catalogue",
            **self._get_request_kwargs(mandatory_key=False),
        )

    @functools.cached_property
    def retrieve_api(self) -> processing.Processing:
        return processing.Processing(
            f"{self.url}/retrieve", **self._get_request_kwargs()
        )

    @functools.cached_property
    def profile_api(self) -> profile.Profile:
        return profile.Profile(f"{self.url}/profiles", **self._get_request_kwargs())

    def check_authentication(self) -> dict[str, Any]:
        return self.profile_api.check_authentication()

    def collections(self, **params: dict[str, Any]) -> catalogue.Collections:
        return self.catalogue_api.collections(params=params)

    def collection(self, collection_id: str) -> catalogue.Collection:
        return self.catalogue_api.collection(collection_id)

    def processes(self, **params: dict[str, Any]) -> processing.ProcessList:
        return self.retrieve_api.processes(params=params)

    def process(self, process_id: str) -> processing.Process:
        return self.retrieve_api.process(process_id=process_id)

    def submit(self, collection_id: str, **request: Any) -> processing.Remote:
        return self.retrieve_api.submit(collection_id, **request)

    def submit_and_wait_on_result(
        self, collection_id: str, **request: Any
    ) -> processing.Results:
        return self.retrieve_api.submit_and_wait_on_result(collection_id, **request)

    def retrieve(
        self,
        collection_id: str,
        target: str | None = None,
        **request: Any,
    ) -> str:
        result = self.submit_and_wait_on_result(collection_id, **request)
        return result.download(target)

    def get_requests(self, **params: dict[str, Any]) -> processing.JobList:
        return self.retrieve_api.jobs(params=params)

    def get_request(self, request_uid: str) -> processing.StatusInfo:
        return self.retrieve_api.job(request_uid)

    def get_remote(self, request_uid: str) -> processing.Remote:
        request = self.get_request(request_uid=request_uid)
        return request.make_remote()

    def download_result(self, request_uid: str, target: str | None) -> str:
        return self.retrieve_api.download_result(request_uid, target)

    def valid_values(
        self, collection_id: str, request: dict[str, Any]
    ) -> dict[str, Any]:
        process = self.retrieve_api.process(collection_id)
        return process.valid_values(request)

    @property
    def licences(self) -> dict[str, Any]:
        return self.catalogue_api.licenses()

    @property
    def accepted_licences(self) -> dict[str, Any]:
        return self.profile_api.accepted_licences()

    def accept_licence(self, licence_id: str, revision: int) -> dict[str, Any]:
        return self.profile_api.accept_licence(licence_id, revision=revision)



import os
from pathlib import Path
from typing import Any

import pytest

from cads_api_client import ApiClient


@pytest.fixture
def api_anon_client(api_root_url: str, api_anon_key: str) -> ApiClient:
    return ApiClient(url=api_root_url, key=api_anon_key, maximum_tries=0)


def test_features_url_cds_adaptor_area_selection(
    tmp_path: Path,
    api_anon_client: ApiClient,
) -> None:
    collection_id = "test-adaptor-url"
    request: dict[str, Any] = {
        "variable": "grid_point_altitude",
        "reference_dataset": "cru",
        "version": "2.1",
    }

    result_bigger = api_anon_client.retrieve(
        collection_id,
        **request,
        target=str(tmp_path / "bigger.zip"),
    )
    result_smaller = api_anon_client.retrieve(
        collection_id,
        **request,
        target=str(tmp_path / "smaller.zip"),
        area=[50, 0, 40, 10],
    )
    assert os.path.getsize(result_bigger) > os.path.getsize(result_smaller)


@pytest.mark.parametrize(
    "format,expected_extension",
    [
        ("grib", ".grib"),
        ("netcdf", ".nc"),
    ],
)
def test_features_mars_cds_adaptor_format(
    api_anon_client: ApiClient,
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
    format: str,
    expected_extension: str,
) -> None:
    monkeypatch.chdir(tmp_path)

    collection_id = "test-adaptor-mars"
    request: dict[str, Any] = {
        "product_type": "reanalysis",
        "variable": "2m_temperature",
        "year": "2016",
        "month": "01",
        "day": "02",
        "time": "00:00",
        "target": None,
    }

    result = api_anon_client.retrieve(
        collection_id,
        **request,
        format=format,
    )

    _, actual_extension = os.path.splitext(result)
    assert actual_extension == expected_extension
    assert os.path.getsize(result)



from __future__ import annotations

import contextlib
import datetime
import os
import pathlib
from typing import Any

import pytest
import requests
from urllib3.exceptions import InsecureRequestWarning

from cads_api_client import ApiClient

does_not_raise = contextlib.nullcontext


@pytest.fixture
def api_anon_client(api_root_url: str, api_anon_key: str) -> ApiClient:
    return ApiClient(url=api_root_url, key=api_anon_key, maximum_tries=0)


def test_accept_licence() -> None:
    client = ApiClient(maximum_tries=0)

    licence = client.licences["licences"][0]
    licence_id = licence["id"]
    licence_revision = licence["revision"]

    expected = {"id": licence_id, "revision": licence_revision}
    actual = client.accept_licence(licence_id, licence_revision)
    assert expected == actual

    assert any(
        licence["id"] == licence_id and licence["revision"] == licence_revision
        for licence in client.accepted_licences["licences"]
    )


def test_delete_request(api_anon_client: ApiClient) -> None:
    remote = api_anon_client.submit(
        "test-adaptor-dummy", _timestamp=datetime.datetime.now().isoformat()
    )
    reply = remote.delete()
    assert reply["status"] == "dismissed"
    with pytest.raises(requests.exceptions.HTTPError):
        remote.status


def test_check_authentication(api_root_url: str, api_anon_client: ApiClient) -> None:
    assert api_anon_client.check_authentication() == {
        "id": -1,
        "role": "anonymous",
        "sub": "anonymous",
    }

    bad_client = ApiClient(key="foo", url=api_root_url)
    with pytest.raises(requests.exceptions.HTTPError, match="401 Client Error"):
        bad_client.check_authentication()


def test_download_result(api_anon_client: ApiClient, tmp_path: pathlib.Path) -> None:
    remote = api_anon_client.submit("test-adaptor-dummy")
    target = str(tmp_path / "test.grib")

    result = api_anon_client.download_result(remote.request_uid, target)
    assert result == target
    assert os.path.exists(result)


def test_get_remote(api_anon_client: ApiClient, tmp_path: pathlib.Path) -> None:
    request_uid = api_anon_client.submit("test-adaptor-dummy").request_uid

    result = api_anon_client.get_remote(request_uid)
    assert result.request_uid == request_uid


def test_api_client_verify(
    api_root_url: str,
    api_anon_key: str,
    tmp_path: pathlib.Path,
) -> None:
    insecure_client = ApiClient(
        url=api_root_url, key=api_anon_key, verify=False, maximum_tries=0
    )
    with pytest.warns(InsecureRequestWarning):
        insecure_client.retrieve(
            "test-adaptor-dummy", target=str(tmp_path / "test.grib")
        )


def test_api_client_timeout(
    api_root_url: str,
    api_anon_key: str,
    tmp_path: pathlib.Path,
) -> None:
    client = ApiClient(url=api_root_url, key=api_anon_key, timeout=0)
    with pytest.raises(ValueError, match="timeout"):
        client.retrieve("test-adaptor-dummy", target=str(tmp_path / "test.grib"))


@pytest.mark.parametrize("progress", [True, False])
def test_api_client_progress(
    api_root_url: str,
    api_anon_key: str,
    tmp_path: pathlib.Path,
    progress: bool,
    capsys: pytest.CaptureFixture[str],
) -> None:
    with capsys.disabled():
        client = ApiClient(
            url=api_root_url, key=api_anon_key, progress=progress, maximum_tries=0
        )
        submitted = client.submit("test-adaptor-dummy")
    submitted.download(target=str(tmp_path / "test.grib"))
    captured = capsys.readouterr()
    assert captured.err if progress else not captured.err


@pytest.mark.parametrize(
    "cleanup,raises",
    [
        (True, pytest.raises(requests.exceptions.HTTPError, match="404 Client Error")),
        (False, does_not_raise()),
    ],
)
def test_api_client_cleanup(
    api_root_url: str,
    api_anon_key: str,
    cleanup: bool,
    raises: contextlib.nullcontext[Any],
) -> None:
    client = ApiClient(
        url=api_root_url, key=api_anon_key, cleanup=cleanup, maximum_tries=0
    )
    remote = client.submit("test-adaptor-dummy")
    request_uid = remote.request_uid
    del remote

    client = ApiClient(url=api_root_url, key=api_anon_key, maximum_tries=0)
    with raises:
        client.get_request(request_uid)



from __future__ import annotations

import json
import os
from typing import Any

SUPPORTED_API_VERSION = "v1"


def read_configuration_file(config_path: str | None = None) -> dict[Any, Any]:
    if config_path is None:
        config_path = os.getenv("CADS_API_RC", "~/.cads-api-client.json")
    config_path = os.path.expanduser(config_path)
    try:
        with open(config_path) as fin:
            config = json.load(fin)
        assert isinstance(config, dict)
    except FileNotFoundError:
        raise
    except Exception:
        raise ValueError(f"failed to parse {config_path!r} file")
    return config


def get_config(key: str, config_path: str | None = None) -> Any:
    return (
        os.getenv(f"CADS_API_{key.upper()}")
        or read_configuration_file(config_path)[key]
    )



from __future__ import annotations

import functools
import logging
import os
import time
import urllib.parse
import warnings
from typing import Any, Type, TypedDict, TypeVar

try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

import attrs
import multiurl
import requests

from . import config

T_ApiResponse = TypeVar("T_ApiResponse", bound="ApiResponse")

logger = logging.getLogger(__name__)


class RequestKwargs(TypedDict):
    headers: dict[str, str]
    session: requests.Session
    retry_options: dict[str, Any]
    request_options: dict[str, Any]
    download_options: dict[str, Any]
    sleep_max: int
    cleanup: bool


class ProcessingFailedError(RuntimeError):
    pass


class DownloadError(RuntimeError):
    pass


class LinkError(Exception):
    pass


def error_json_to_message(error_json: dict[str, Any]) -> str:
    error_messages = [
        str(error_json[key])
        for key in ("title", "traceback", "detail")
        if key in error_json
    ]
    return "\n".join(error_messages)


def cads_raise_for_status(response: requests.Response) -> None:
    if 400 <= response.status_code < 500:
        try:
            error_json = response.json()
        except Exception:
            pass
        else:
            message = "\n".join(
                [
                    f"{response.status_code} Client Error: {response.reason} for url: {response.url}",
                    error_json_to_message(error_json),
                ]
            )
            raise requests.exceptions.HTTPError(message, response=response)
    response.raise_for_status()


def get_level_and_message(message: str) -> tuple[int, str]:
    level = 20
    for severity in ("DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"):
        if message.startswith(severity):
            level = logging.getLevelName(severity)
            message = message.replace(severity, "", 1).lstrip(":").lstrip()
            break
    return level, message


@attrs.define(slots=False)
class ApiResponse:
    response: requests.Response
    headers: dict[str, str]
    session: requests.Session
    retry_options: dict[str, Any]
    request_options: dict[str, Any]
    download_options: dict[str, Any]
    sleep_max: int
    cleanup: bool

    @property
    def request_kwargs(self) -> RequestKwargs:
        return RequestKwargs(
            headers=self.headers,
            session=self.session,
            retry_options=self.retry_options,
            request_options=self.request_options,
            download_options=self.download_options,
            sleep_max=self.sleep_max,
            cleanup=self.cleanup,
        )

    @classmethod
    def from_request(
        cls: Type[T_ApiResponse],
        method: str,
        url: str,
        headers: dict[str, str],
        session: requests.Session | None,
        retry_options: dict[str, Any],
        request_options: dict[str, Any],
        download_options: dict[str, Any],
        sleep_max: int,
        cleanup: bool,
        **kwargs: Any,
    ) -> T_ApiResponse:
        if session is None:
            session = requests.Session()
        robust_request = multiurl.robust(session.request, **retry_options)

        inputs = kwargs.get("json", {}).get("inputs", {})
        logger.debug(f"{method.upper()} {url} {inputs or ''}".strip())
        response = robust_request(
            method, url, headers=headers, **request_options, **kwargs
        )
        logger.debug(f"REPLY {response.text}")

        cads_raise_for_status(response)

        self = cls(
            response,
            headers=headers,
            session=session,
            retry_options=retry_options,
            request_options=request_options,
            download_options=download_options,
            sleep_max=sleep_max,
            cleanup=cleanup,
        )
        self.log_messages()
        return self

    @property
    def json(self) -> dict[str, Any]:
        json: dict[str, Any] = self.response.json()
        return json

    def log_messages(self) -> None:
        if message := self.json.get("message"):
            level, message = get_level_and_message(message)
            logger.log(level, message)

        dataset_messages = (
            self.json.get("metadata", {}).get("datasetMetadata", {}).get("messages", [])
        )
        for dataset_message in dataset_messages:
            if not (content := dataset_message.get("content")):
                continue
            if date := dataset_message.get("date"):
                content = f"[{date}] {content}"
            severity = dataset_message.get("severity", "notset").upper()
            level = logging.getLevelName(severity)
            logger.log(level if isinstance(level, int) else 20, content)

    def get_links(self, rel: str | None = None) -> list[dict[str, str]]:
        links = []
        for link in self.json.get("links", []):
            if rel is not None and link.get("rel") == rel:
                links.append(link)
        return links

    def get_link_href(self, rel: str | None = None) -> str:
        links = self.get_links(rel)
        if len(links) != 1:
            raise LinkError(f"link not found or not unique {rel=}")
        return links[0]["href"]

    def from_rel_href(self, rel: str) -> Self | None:
        rels = self.get_links(rel=rel)
        if len(rels) > 1:
            raise LinkError(f"link not unique {rel=}")

        if len(rels) == 1:
            out = self.from_request("get", rels[0]["href"], **self.request_kwargs)
        else:
            out = None
        return out


@attrs.define
class ProcessList(ApiResponse):
    def process_ids(self) -> list[str]:
        return [proc["id"] for proc in self.json["processes"]]

    def next(self) -> ApiResponse | None:
        return self.from_rel_href(rel="next")

    def prev(self) -> ApiResponse | None:
        return self.from_rel_href(rel="prev")


@attrs.define
class Process(ApiResponse):
    @property
    def id(self) -> str:
        process_id: str = self.json["id"]
        return process_id

    def execute(self, inputs: dict[str, Any]) -> StatusInfo:
        url = f"{self.response.request.url}/execution"
        return StatusInfo.from_request(
            "post", url, json={"inputs": inputs}, **self.request_kwargs
        )

    def valid_values(self, request: dict[str, Any] = {}) -> dict[str, Any]:
        url = f"{self.response.request.url}/constraints"
        response = ApiResponse.from_request(
            "post", url, json={"inputs": request}, **self.request_kwargs
        )
        response.response.raise_for_status()
        return response.json


@attrs.define(slots=False)
class Remote:
    url: str
    headers: dict[str, str]
    session: requests.Session
    retry_options: dict[str, Any]
    request_options: dict[str, Any]
    download_options: dict[str, Any]
    sleep_max: int
    cleanup: bool

    def __attrs_post_init__(self) -> None:
        self.log_start_time = None
        self.info(f"Request ID is {self.request_uid}")

    @property
    def request_kwargs(self) -> RequestKwargs:
        return RequestKwargs(
            headers=self.headers,
            session=self.session,
            retry_options=self.retry_options,
            request_options=self.request_options,
            download_options=self.download_options,
            sleep_max=self.sleep_max,
            cleanup=self.cleanup,
        )

    def log_metadata(self, metadata: dict[str, Any]) -> None:
        logs = metadata.get("log", [])
        for self.log_start_time, message in sorted(logs):
            level, message = get_level_and_message(message)
            logger.log(level, message)

    def get_api_response(self, method: str, **kwargs: Any) -> ApiResponse:
        return ApiResponse.from_request(
            method, self.url, **self.request_kwargs, **kwargs
        )

    @functools.cached_property
    def request_uid(self) -> str:
        return self.url.rpartition("/")[2]

    @property
    def _reply(self) -> dict[str, Any]:
        params = {"log": True}
        if self.log_start_time:
            params["logStartTime"] = self.log_start_time
        return self.get_api_response("get", params=params).json

    @property
    def status(self) -> str:
        reply = self._reply
        self.log_metadata(reply.get("metadata", {}))
        status: str = reply["status"]
        return status

    def wait_on_result(self) -> None:
        sleep = 1.0
        status = None
        while True:
            if status != (status := self.status):
                self.info(f"status has been updated to {status}")
            if status == "successful":
                break
            elif status == "failed":
                results = self.make_results()
                raise ProcessingFailedError(error_json_to_message(results.json))
            elif status in ("accepted", "running"):
                sleep *= 1.5
                if sleep > self.sleep_max:
                    sleep = self.sleep_max
            elif status in ("dismissed", "deleted"):
                raise ProcessingFailedError(f"API state {status!r}")
            else:
                raise ProcessingFailedError(f"Unknown API state {status!r}")
            self.debug(f"result not ready, waiting for {sleep} seconds")
            time.sleep(sleep)

    def build_status_info(self) -> StatusInfo:
        return StatusInfo.from_request("get", self.url, **self.request_kwargs)

    def make_results(self) -> Results:
        response = self.get_api_response("get")
        try:
            results_url = response.get_link_href(rel="results")
        except LinkError:
            results_url = f"{self.url}/results"
        results = Results.from_request("get", results_url, **self.request_kwargs)
        return results

    def _download_result(
        self,
        target: str | None = None,
    ) -> str:
        results = self.make_results()
        return results.download(target)

    def download(
        self,
        target: str | None = None,
    ) -> str:
        self.wait_on_result()
        return self._download_result(target)

    def delete(self) -> dict[str, Any]:
        response = self.get_api_response("delete")
        self.cleanup = False
        return response.json

    def _warn(self) -> None:
        message = (
            ".update and .reply are available for backward compatibility."
            " You can now use .download directly without needing to check whether the request is completed."
        )
        warnings.warn(message, DeprecationWarning)

    def update(self, request_id: str | None = None) -> None:
        self._warn()
        if request_id:
            assert request_id == self.request_uid
        try:
            del self.reply
        except AttributeError:
            pass
        self.reply

    @functools.cached_property
    def reply(self) -> dict[str, Any]:
        self._warn()

        reply = self._reply
        reply.setdefault("state", reply["status"])

        if reply["state"] == "successful":
            reply["state"] = "completed"
        elif reply["state"] == "queued":
            reply["state"] = "accepted"
        elif reply["state"] == "failed":
            reply.setdefault("error", {})
            try:
                self.make_results()
            except Exception as exc:
                reply["error"].setdefault("message", str(exc))

        reply.setdefault("request_id", self.request_uid)
        return reply

    def info(self, *args: Any, **kwargs: Any) -> None:
        logger.info(*args, **kwargs)

    def warning(self, *args: Any, **kwargs: Any) -> None:
        logger.warning(*args, **kwargs)

    def error(self, *args: Any, **kwargs: Any) -> None:
        logger.error(*args, **kwargs)

    def debug(self, *args: Any, **kwargs: Any) -> None:
        logger.debug(*args, **kwargs)

    def __del__(self) -> None:
        if self.cleanup:
            try:
                self.delete()
            except Exception as exc:
                warnings.warn(str(exc), UserWarning)


@attrs.define
class StatusInfo(ApiResponse):
    def make_remote(self) -> Remote:
        if self.response.request.method == "POST":
            url = self.get_link_href(rel="monitor")
        else:
            url = self.get_link_href(rel="self")
        return Remote(url, **self.request_kwargs)


@attrs.define
class JobList(ApiResponse):
    def job_ids(self) -> list[str]:
        return [job["jobID"] for job in self.json["jobs"]]

    def next(self) -> ApiResponse | None:
        return self.from_rel_href(rel="next")

    def prev(self) -> ApiResponse | None:
        return self.from_rel_href(rel="prev")


@attrs.define
class Results(ApiResponse):
    @property
    def status_code(self) -> int:
        return self.response.status_code

    @property
    def reason(self) -> str:
        return self.response.reason

    def get_result_href(self) -> str:
        if self.status_code != 200:
            raise KeyError("result_href not available for processing failed results")
        href: str = self.json["asset"]["value"]["href"]
        return href

    @property
    def asset(self) -> dict[str, Any]:
        return dict(self.json["asset"]["value"])

    @property
    def location(self) -> str:
        result_href = self.get_result_href()
        return urllib.parse.urljoin(self.response.url, result_href)

    @property
    def content_length(self) -> int:
        return int(self.asset["file:size"])

    @property
    def content_type(self) -> str:
        return str(self.asset["type"])

    def download(
        self,
        target: str | None = None,
    ) -> str:
        url = self.location
        if target is None:
            parts = urllib.parse.urlparse(url)
            target = parts.path.strip("/").split("/")[-1]

        download_options = {"stream": True}
        download_options.update(self.download_options)
        multiurl.download(
            url,
            target=target,
            **self.retry_options,
            **self.request_options,
            **download_options,
        )
        if (target_size := os.path.getsize(target)) != (size := self.content_length):
            raise DownloadError(
                "Download failed: downloaded %s byte(s) out of %s" % (target_size, size)
            )
        return target

    def info(self, *args: Any, **kwargs: Any) -> None:
        logger.info(*args, **kwargs)

    def warning(self, *args: Any, **kwargs: Any) -> None:
        logger.warning(*args, **kwargs)

    def error(self, *args: Any, **kwargs: Any) -> None:
        logger.error(*args, **kwargs)

    def debug(self, *args: Any, **kwargs: Any) -> None:
        logger.debug(*args, **kwargs)


@attrs.define(slots=False)
class Processing:
    url: str
    headers: dict[str, str]
    session: requests.Session
    retry_options: dict[str, Any]
    request_options: dict[str, Any]
    download_options: dict[str, Any]
    sleep_max: int
    cleanup: bool
    force_exact_url: bool = False

    def __attrs_post_init__(self) -> None:
        if not self.force_exact_url:
            self.url += f"/{config.SUPPORTED_API_VERSION}"

    @property
    def request_kwargs(self) -> RequestKwargs:
        return RequestKwargs(
            headers=self.headers,
            session=self.session,
            retry_options=self.retry_options,
            request_options=self.request_options,
            download_options=self.download_options,
            sleep_max=self.sleep_max,
            cleanup=self.cleanup,
        )

    def processes(self, params: dict[str, Any] = {}) -> ProcessList:
        url = f"{self.url}/processes"
        return ProcessList.from_request(
            "get", url, params=params, **self.request_kwargs
        )

    def process(self, process_id: str) -> Process:
        url = f"{self.url}/processes/{process_id}"
        return Process.from_request("get", url, **self.request_kwargs)

    def process_execute(
        self,
        process_id: str,
        inputs: dict[str, Any],
    ) -> StatusInfo:
        url = f"{self.url}/processes/{process_id}/execution"
        return StatusInfo.from_request(
            "post", url, json={"inputs": inputs}, **self.request_kwargs
        )

    def jobs(self, params: dict[str, Any] = {}) -> JobList:
        url = f"{self.url}/jobs"
        return JobList.from_request("get", url, params=params, **self.request_kwargs)

    def job(self, job_id: str) -> StatusInfo:
        url = f"{self.url}/jobs/{job_id}"
        return StatusInfo.from_request("get", url, **self.request_kwargs)

    def job_results(self, job_id: str) -> Results:
        url = f"{self.url}/jobs/{job_id}/results"
        return Results.from_request("get", url, **self.request_kwargs)

    # convenience methods

    def submit(self, collection_id: str, **request: Any) -> Remote:
        status_info = self.process_execute(collection_id, request)
        return status_info.make_remote()

    def submit_and_wait_on_result(self, collection_id: str, **request: Any) -> Results:
        remote = self.submit(collection_id, **request)
        remote.wait_on_result()
        return remote.make_results()

    def make_remote(self, job_id: str) -> Remote:
        url = f"{self.url}/jobs/{job_id}"
        return Remote(url, **self.request_kwargs)

    def download_result(self, job_id: str, target: str | None) -> str:
        # NOTE: the remote waits for the result to be available
        return self.make_remote(job_id).download(target)



from __future__ import annotations

import functools
import logging
import warnings
from types import TracebackType
from typing import Any, Callable, TypeVar, cast, overload

import cdsapi.api
import requests

from . import __version__ as cads_api_client_version
from . import api_client, processing

LEGACY_KWARGS = [
    "full_stack",
    "delete",
    "retry_max",
    "sleep_max",
    "wait_until_complete",
    "info_callback",
    "warning_callback",
    "error_callback",
    "debug_callback",
    "metadata",
    "forget",
    "session",
]

LOGGER = logging.getLogger(__name__)
F = TypeVar("F", bound=Callable[..., Any])


class LoggingContext:
    def __init__(self, logger: logging.Logger, quiet: bool, debug: bool) -> None:
        self.old_level = logger.level
        if quiet:
            logger.setLevel(logging.WARNING)
        else:
            logger.setLevel(logging.DEBUG if debug else logging.INFO)

        self.new_handlers = []
        if not logger.handlers:
            formatter = logging.Formatter("%(asctime)s %(levelname)s %(message)s")
            handler = logging.StreamHandler()
            handler.setFormatter(formatter)
            logger.addHandler(handler)
            self.new_handlers.append(handler)

        self.logger = logger

    def __enter__(self) -> logging.Logger:
        return self.logger

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self.logger.setLevel(self.old_level)
        for handler in self.new_handlers:
            self.logger.removeHandler(handler)


class LegacyApiClient(cdsapi.api.Client):  # type: ignore[misc]
    def __init__(
        self,
        url: str | None = None,
        key: str | None = None,
        quiet: bool = False,
        debug: bool = False,
        verify: bool | None = None,
        timeout: int = 60,
        progress: bool = True,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        kwargs.update(zip(LEGACY_KWARGS, args))
        if wrong_kwargs := set(kwargs) - set(LEGACY_KWARGS):
            raise ValueError(f"Wrong parameters: {wrong_kwargs}.")

        self.url, self.key, self.verify = cdsapi.api.get_url_key_verify(
            url, key, verify
        )
        self.quiet = quiet
        self._debug = debug
        self.timeout = timeout
        self.progress = progress

        self.sleep_max = kwargs.pop("sleep_max", 120)
        self.wait_until_complete = kwargs.pop("wait_until_complete", True)
        self.delete = kwargs.pop("delete", False)
        self.retry_max = kwargs.pop("retry_max", 500)
        self.session = kwargs.pop("session", requests.Session())
        if kwargs:
            warnings.warn(
                "This is a beta version."
                f" The following parameters have not been implemented yet: {kwargs}.",
                UserWarning,
            )

        self.client = api_client.ApiClient(
            url=self.url,
            key=self.key,
            verify=self.verify,
            sleep_max=self.sleep_max,
            session=self.session,
            cleanup=self.delete,
            maximum_tries=self.retry_max,
            retry_after=self.sleep_max,
            timeout=self.timeout,
            progress=self.progress,
        )
        self.debug(
            "CDSAPI %s",
            {
                "url": self.url,
                "key": self.key,
                "quiet": self.quiet,
                "verify": self.verify,
                "timeout": self.timeout,
                "progress": self.progress,
                "sleep_max": self.sleep_max,
                "retry_max": self.retry_max,
                "delete": self.delete,
                "cads_api_client_version": cads_api_client_version,
            },
        )

    @classmethod
    def raise_not_implemented_error(self) -> None:
        raise NotImplementedError(
            "This is a beta version. This functionality has not been implemented yet."
        )

    def logging_decorator(self, func: F) -> F:
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            with LoggingContext(
                logger=processing.logger, quiet=self.quiet, debug=self._debug
            ):
                return func(*args, **kwargs)

        return cast(F, wrapper)

    @overload
    def retrieve(self, name: str, request: dict[str, Any], target: str) -> str: ...

    @overload
    def retrieve(
        self, name: str, request: dict[str, Any], target: None = ...
    ) -> processing.Results: ...

    def retrieve(
        self, name: str, request: dict[str, Any], target: str | None = None
    ) -> str | processing.Remote | processing.Results:
        submitted: processing.Remote | processing.Results
        if self.wait_until_complete:
            submitted = self.logging_decorator(self.client.submit_and_wait_on_result)(
                collection_id=name,
                **request,
            )
        else:
            submitted = self.logging_decorator(self.client.submit)(
                collection_id=name,
                **request,
            )

        # Decorate legacy methods
        submitted.download = self.logging_decorator(submitted.download)  # type: ignore[method-assign]
        submitted.info = self.logging_decorator(submitted.info)  # type: ignore[method-assign]
        submitted.warning = self.logging_decorator(submitted.warning)  # type: ignore[method-assign]
        submitted.error = self.logging_decorator(submitted.error)  # type: ignore[method-assign]
        submitted.debug = self.logging_decorator(submitted.debug)  # type: ignore[method-assign]

        return submitted if target is None else submitted.download(target)

    def info(self, *args: Any, **kwargs: Any) -> None:
        with LoggingContext(
            logger=LOGGER, quiet=self.quiet, debug=self._debug
        ) as logger:
            logger.info(*args, **kwargs)

    def warning(self, *args: Any, **kwargs: Any) -> None:
        with LoggingContext(
            logger=LOGGER, quiet=self.quiet, debug=self._debug
        ) as logger:
            logger.warning(*args, **kwargs)

    def error(self, *args: Any, **kwargs: Any) -> None:
        with LoggingContext(
            logger=LOGGER, quiet=self.quiet, debug=self._debug
        ) as logger:
            logger.error(*args, **kwargs)

    def debug(self, *args: Any, **kwargs: Any) -> None:
        with LoggingContext(
            logger=LOGGER, quiet=self.quiet, debug=self._debug
        ) as logger:
            logger.debug(*args, **kwargs)

    def service(self, name, *args, **kwargs):  # type: ignore
        self.raise_not_implemented_error()

    def workflow(self, code, *args, **kwargs):  # type: ignore
        self.raise_not_implemented_error()

    def status(self, context=None):  # type: ignore
        self.raise_not_implemented_error()

    def download(self, results, targets=None):  # type: ignore
        self.raise_not_implemented_error()

    def remote(self, url):  # type: ignore
        self.raise_not_implemented_error()

    def robust(self, call):  # type: ignore
        self.raise_not_implemented_error()



import pathlib

import pytest

from cads_api_client import ApiClient, catalogue, processing


@pytest.fixture
def cat(api_root_url: str, api_anon_key: str) -> catalogue.Catalogue:
    client = ApiClient(url=api_root_url, key=api_anon_key, maximum_tries=0)
    return client.catalogue_api


def test_from_collection_to_process(cat: catalogue.Catalogue) -> None:
    collection_id = "test-adaptor-dummy"
    dataset = cat.collection(collection_id)

    res = dataset.retrieve_process()

    assert isinstance(res, processing.Process)


def test_collection_submit(cat: catalogue.Catalogue) -> None:
    collection_id = "test-adaptor-dummy"
    dataset = cat.collection(collection_id)

    res = dataset.submit()

    assert isinstance(res, processing.Remote)

    assert isinstance(res.request_uid, str)
    assert isinstance(res.status, str)


def test_collection_retrieve_with_dummy_adaptor(
    cat: catalogue.Catalogue, tmp_path: pathlib.Path
) -> None:
    collection_id = "test-adaptor-dummy"
    dataset = cat.collection(collection_id)
    target = str(tmp_path / "dummy.txt")

    res = dataset.retrieve(
        target=target,
    )

    assert isinstance(res, str)
    assert res.endswith(target)


def test_collection_retrieve_with_url_cds_adaptor(
    cat: catalogue.Catalogue, tmp_path: pathlib.Path
) -> None:
    collection_id = "test-adaptor-url"
    dataset = cat.collection(collection_id)
    target = str(tmp_path / "wfde1.zip")

    res = dataset.retrieve(
        variable="grid_point_altitude",
        reference_dataset="cru",
        version="2.1",
        target=target,
    )

    assert isinstance(res, str)
    assert res.endswith(target)

    target = str(tmp_path / "wfde2.zip")

    res = dataset.retrieve(
        variable="grid_point_altitude",
        reference_dataset="cru",
        version="2.1",
        format="zip",
        target=target,
    )

    assert isinstance(res, str)
    assert res.endswith(target)


def test_collection_retrieve_with_direct_mars_cds_adaptor(
    cat: catalogue.Catalogue, tmp_path: pathlib.Path
) -> None:
    collection_id = "test-adaptor-direct-mars"
    dataset = cat.collection(collection_id)
    target = str(tmp_path / "era5-complete.grib")

    request = {
        "levelist": "1",
        "dataset": "reanalysis",
        "time": "00:00:00",
        "param": "155",
        "date": "1940-01-01",
        "expect": "any",
        "levtype": "pl",
        "number": "all",
        "class": "ea",
    }
    res = dataset.retrieve(target=target, **request)

    assert isinstance(res, str)
    assert res.endswith(target)


def test_collection_retrieve_with_mars_cds_adaptor(
    cat: catalogue.Catalogue, tmp_path: pathlib.Path
) -> None:
    collection_id = "test-adaptor-mars"
    dataset = cat.collection(collection_id)
    target = str(tmp_path / "era5.grib")

    res = dataset.retrieve(
        product_type="reanalysis",
        variable="2m_temperature",
        year="2016",
        month="01",
        day="02",
        time="00:00",
        target=target,
    )

    assert isinstance(res, str)
    assert res.endswith(target)


@pytest.mark.skip(reason="discontinued adaptor")
def test_collection_retrieve_with_legacy_cds_adaptor(
    cat: catalogue.Catalogue, tmp_path: pathlib.Path
) -> None:
    collection_id = "test-adaptor-legacy"
    dataset = cat.collection(collection_id)
    target = str(tmp_path / "era5.grib")

    res = dataset.retrieve(
        product_type="reanalysis",
        variable="temperature",
        year="2016",
        month="01",
        day="02",
        time="00:00",
        level="1000",
        target=target,
        retry_options={"maximum_tries": 0},
    )

    assert isinstance(res, str)
    assert res.endswith(target)



import json
import logging

import pytest
import requests
import responses
from responses.matchers import json_params_matcher

import cads_api_client

COLLECTION_ID = "reanalysis-era5-pressure-levels"
JOB_RUNNING_ID = "9bfc1362-2832-48e1-a235-359267420bb1"
JOB_SUCCESSFUL_ID = "9bfc1362-2832-48e1-a235-359267420bb2"
JOB_FAILED_ID = "9bfc1362-2832-48e1-a235-359267420bb3"

CATALOGUE_URL = "http://localhost:8080/api/catalogue"
COLLECTIONS_URL = "http://localhost:8080/api/catalogue/v1/datasets"
COLLECTION_URL = (
    "http://localhost:8080/api/catalogue/v1/collections/reanalysis-era5-pressure-levels"
)
PROCESS_URL = f"http://localhost:8080/api/retrieve/v1/processes/{COLLECTION_ID}"
EXECUTE_URL = f"{PROCESS_URL}/execution"

JOB_RUNNING_URL = f"http://localhost:8080/api/retrieve/v1/jobs/{JOB_RUNNING_ID}"
JOB_SUCCESSFUL_URL = f"http://localhost:8080/api/retrieve/v1/jobs/{JOB_SUCCESSFUL_ID}"
JOB_FAILED_URL = f"http://localhost:8080/api/retrieve/v1/jobs/{JOB_FAILED_ID}"

RESULT_RUNNING_URL = (
    f"http://localhost:8080/api/retrieve/v1/jobs/{JOB_RUNNING_ID}/results"
)
RESULT_SUCCESSFUL_URL = (
    f"http://localhost:8080/api/retrieve/v1/jobs/{JOB_SUCCESSFUL_ID}/results"
)
RESULT_FAILED_URL = (
    f"http://localhost:8080/api/retrieve/v1/jobs/{JOB_FAILED_ID}/results"
)


CATALOGUE_JSON = {
    "type": "Catalog",
    "id": "stac-fastapi",
    "links": [
        {"rel": "self", "type": "application/json", "href": f"{CATALOGUE_URL}/v1"},
        {"rel": "root", "type": "application/json", "href": f"{CATALOGUE_URL}/v1"},
        {"rel": "data", "type": "application/json", "href": f"{COLLECTIONS_URL}"},
        {
            "rel": "child",
            "type": "application/json",
            "title": "ERA5 hourly data on pressure levels from 1959 to present",
            "href": f"{COLLECTION_URL}",
        },
    ],
}

COLLECTIONS_JSON = {
    "collections": [
        {
            "type": "Collection",
            "id": f"{COLLECTION_ID}",
            "links": [
                {
                    "rel": "self",
                    "type": "application/json",
                    "href": f"{COLLECTION_URL}",
                },
                {
                    "rel": "parent",
                    "type": "application/json",
                    "href": f"{CATALOGUE_URL}/v1",
                },
                {
                    "rel": "root",
                    "type": "application/json",
                    "href": f"{CATALOGUE_URL}/v1",
                },
            ],
        },
    ],
    "links": [
        {"rel": "root", "type": "application/json", "href": f"{CATALOGUE_URL}/v1"},
        {"rel": "parent", "type": "application/json", "href": f"{CATALOGUE_URL}/v1"},
        {"rel": "self", "type": "application/json", "href": f"{COLLECTIONS_URL}"},
    ],
}

COLLECTION_JSON = {
    "type": "Collection",
    "id": COLLECTION_ID,
    "links": [
        {
            "rel": "self",
            "type": "application/json",
            "href": "http://localhost:8080/api/catalogue/v1/collections/reanalysis-era5-pressure-levels",
        },
        {
            "rel": "parent",
            "type": "application/json",
            "href": "http://localhost:8080/api/catalogue/v1/",
        },
        {
            "rel": "root",
            "type": "application/json",
            "href": "http://localhost:8080/api/catalogue/v1/",
        },
        {
            "rel": "retrieve",
            "href": "http://localhost:8080/api/retrieve/v1/processes/reanalysis-era5-pressure-levels",
            "type": "application/json",
        },
        {
            "rel": "related",
            "href": "http://localhost:8080/api/catalogue/v1/collections/reanalysis-era5-single-levels",
        },
    ],
    "tmp:variables": {
        "Temperature": {
            "units": "K",
        }
    },
}

PROCESS_JSON = {
    "id": COLLECTION_ID,
    "links": [
        {"href": COLLECTION_URL, "rel": "self", "type": "application/json"},
        {"rel": "retrieve", "href": PROCESS_URL, "type": "application/json"},
        {
            "href": f"{COLLECTION_URL}/execution",
            "rel": "execute",
            "type": "application/json",
            "title": "process execution",
        },
    ],
    "inputs": {
        "product_type": {
            "title": "Product type",
            "schema": {
                "type": "array",
                "items": {"enum": ["ensemble_mean", "reanalysis"], "type": "string"},
            },
        },
        "variable": {
            "title": "Variable",
            "schema": {
                "type": "array",
                "items": {"enum": ["temperature", "vorticity"], "type": "string"},
            },
        },
        "year": {
            "title": "Year",
            "schema": {
                "type": "array",
                "items": {"enum": ["2022", "0000"], "type": "string"},
            },
        },
    },
    "outputs": {
        "download_url": {
            "schema": {"type": "string", "format": "url"},
        }
    },
    "message": "WARNING: This is a warning message",
    "metadata": {
        "datasetMetadata": {
            "messages": [
                {
                    "date": "2023-12-12T13:00:00",
                    "severity": "warning",
                    "content": "This is a warning dataset message",
                },
                {
                    "date": "2023-12-12T14:00:00",
                    "severity": "success",
                    "content": "This is a success dataset message",
                },
            ]
        }
    },
}

JOB_RUNNING_JSON = {
    "processID": f"{COLLECTION_ID}",
    "type": "process",
    "jobID": f"{JOB_RUNNING_ID}",
    "status": "running",
    "created": "2022-09-02T17:30:48.201213",
    "updated": "2022-09-02T17:30:48.201217",
    "links": [
        {
            "href": f"{COLLECTION_URL}/execution",
            "rel": "self",
            "type": "application/json",
        },
        {
            "href": f"{JOB_RUNNING_URL}",
            "rel": "monitor",
            "type": "application/json",
            "title": "job status info",
        },
    ],
}


JOB_SUCCESSFUL_JSON = {
    "processID": f"{COLLECTION_ID}",
    "type": "process",
    "jobID": f"{JOB_SUCCESSFUL_ID}",
    "status": "successful",
    "created": "2022-09-02T17:30:48.201213",
    "started": "2022-09-02T17:32:43.890617",
    "finished": "2022-09-02T17:32:54.308120",
    "updated": "2022-09-02T17:32:54.308116",
    "links": [
        {"href": f"{JOB_SUCCESSFUL_ID}", "rel": "self", "type": "application/json"},
        {
            "href": f"http://localhost:8080/api/retrieve/v1/jobs/{JOB_SUCCESSFUL_ID}/results",
            "rel": "results",
        },
        {
            "href": f"{JOB_SUCCESSFUL_URL}",
            "rel": "monitor",
            "type": "application/json",
            "title": "job status info",
        },
    ],
    "metadata": {
        "log": [
            ["2024-02-09T09:14:47.811223", "This is a log"],
            ["2024-02-09T09:14:50.811223", "WARNING: This is a warning log"],
        ]
    },
}

JOB_FAILED_JSON = {
    "processID": f"{COLLECTION_ID}",
    "type": "process",
    "jobID": f"{JOB_FAILED_ID}",
    "status": "failed",
    "created": "2022-09-02T17:30:48.201213",
    "started": "2022-09-02T17:32:43.890617",
    "finished": "2022-09-02T17:32:54.308120",
    "updated": "2022-09-02T17:32:54.308116",
    "links": [
        {"href": f"{JOB_FAILED_ID}", "rel": "self", "type": "application/json"},
        {
            "href": f"http://localhost:8080/api/retrieve/v1/jobs/{JOB_FAILED_ID}/results",
            "rel": "results",
        },
        {
            "href": f"{JOB_FAILED_URL}",
            "rel": "monitor",
            "type": "application/json",
            "title": "job status info",
        },
    ],
    "metadata": {
        "log": [
            ["2024-02-09T09:14:47.811223", "This is a log"],
            ["2024-02-09T09:14:50.811223", "WARNING: This is a warning log"],
        ]
    },
}

RESULT_SUCCESSFUL_JSON = {
    "asset": {
        "value": {
            "type": "application/netcdf",
            "href": "./e7d452a747061ab880887d88814bfb0c27593a73cb7736d2dc340852",
            "file:checksum": "e7d452a747061ab880887d88814bfb0c27593a73cb7736d2dc340852",
            "file:size": 8,
            "file:local_path": [
                "/cache-store/",
                "e7d452a747061ab880887d88814bfb0c27593a73cb7736d2dc340852.nc",
            ],
            "xarray:open_kwargs": {},
            "xarray:storage_options": {},
        }
    }
}


RESULT_RUNNING_JSON = {
    "type": "http://www.opengis.net/def/exceptions/ogcapi-processes-1/1.0/result-not-ready",
    "title": "job results not ready",
    "detail": "job 8b7a1f3d-04b1-425d-96f1-f0634d02ee7f results are not yet ready",
    "instance": "http://127.0.0.1:8080/api/retrieve/v1/jobs/8b7a1f3d-04b1-425d-96f1-f0634d02ee7f/results",
}

RESULT_FAILED_JSON = {
    "type": "job results failed",
    "title": "job failed",
    "status": 400,
    "instance": "http://127.0.0.1:8080/api/retrieve/v1/jobs/02135eee-39a8-4d1f-8cd7-87682de5b981/results",
    "trace_id": "ca3e7170-1ce2-48fc-97f8-bbe64fafce44",
    "traceback": "This is a traceback",
}


@pytest.fixture
def catalogue() -> cads_api_client.Catalogue:
    return cads_api_client.Catalogue(
        CATALOGUE_URL,
        headers={},
        session=requests.Session(),
        retry_options={},
        request_options={},
        download_options={},
        sleep_max=120,
        cleanup=False,
    )


def responses_add() -> None:
    responses.add(
        responses.GET,
        url=f"{CATALOGUE_URL}/v1/",
        json=CATALOGUE_JSON,
        content_type="application/json",
    )

    responses.add(
        responses.GET,
        url=f"{COLLECTIONS_URL}",
        json=COLLECTIONS_JSON,
        content_type="application/json",
    )

    responses.add(
        responses.GET,
        url=f"{COLLECTION_URL}",
        json=COLLECTION_JSON,
        content_type="application/json",
    )

    responses.add(
        responses.GET,
        url=PROCESS_URL,
        json=PROCESS_JSON,
        content_type="application/json",
    )

    responses.add(
        responses.POST,
        url=EXECUTE_URL,
        json=JOB_SUCCESSFUL_JSON,
        match=[
            json_params_matcher(
                {
                    "inputs": {"variable": "temperature", "year": "2022"},
                }
            )
        ],
        content_type="application/json",
    )

    responses.add(
        responses.GET,
        url=JOB_SUCCESSFUL_URL,
        json=JOB_SUCCESSFUL_JSON,
        content_type="application/json",
    )

    responses.add(
        responses.POST,
        url=EXECUTE_URL,
        json=JOB_FAILED_JSON,
        match=[
            json_params_matcher(
                {
                    "inputs": {"variable": "temperature", "year": "0000"},
                }
            )
        ],
        content_type="application/json",
    )

    responses.add(
        responses.GET,
        url=JOB_FAILED_URL,
        json=JOB_FAILED_JSON,
        content_type="application/json",
    )

    responses.add(
        responses.GET,
        url=RESULT_FAILED_URL,
        json=RESULT_FAILED_JSON,
        content_type="application/json",
    )


@responses.activate
def test_catalogue_collections(catalogue: cads_api_client.Catalogue) -> None:
    responses_add()

    collections = catalogue.collections()
    assert collections.response.json() == COLLECTIONS_JSON

    collection = catalogue.collection(COLLECTION_ID)
    assert collection.response.json() == COLLECTION_JSON


@responses.activate
def test_submit(catalogue: cads_api_client.Catalogue) -> None:
    responses_add()

    collection = catalogue.collection(COLLECTION_ID)
    process = collection.retrieve_process()

    assert process.response.json() == PROCESS_JSON

    job = process.execute(inputs={"variable": "temperature", "year": "2022"})
    assert job.response.json() == JOB_SUCCESSFUL_JSON

    remote = collection.submit(variable="temperature", year="2022")
    assert remote.url == JOB_SUCCESSFUL_URL
    assert remote.status == "successful"


@responses.activate
def test_wait_on_result(catalogue: cads_api_client.Catalogue) -> None:
    responses_add()

    collection = catalogue.collection(COLLECTION_ID)
    remote = collection.submit(variable="temperature", year="2022")
    remote.wait_on_result()


@responses.activate
def test_wait_on_result_failed(catalogue: cads_api_client.Catalogue) -> None:
    responses_add()

    collection = catalogue.collection(COLLECTION_ID)
    remote = collection.submit(variable="temperature", year="0000")
    with pytest.raises(
        cads_api_client.processing.ProcessingFailedError,
        match="job failed\nThis is a traceback",
    ):
        remote.wait_on_result()


@responses.activate
def test_remote_logs(
    caplog: pytest.LogCaptureFixture, catalogue: cads_api_client.Catalogue
) -> None:
    responses_add()

    collection = catalogue.collection(COLLECTION_ID)

    with caplog.at_level(logging.DEBUG, logger="cads_api_client.processing"):
        remote = collection.submit(variable="temperature", year="2022")
        remote.wait_on_result()

    assert caplog.record_tuples == [
        (
            "cads_api_client.processing",
            10,
            "GET http://localhost:8080/api/retrieve/v1/processes/reanalysis-era5-pressure-levels",
        ),
        (
            "cads_api_client.processing",
            10,
            f"REPLY {json.dumps(PROCESS_JSON)}",
        ),
        (
            "cads_api_client.processing",
            30,
            "This is a warning message",
        ),
        (
            "cads_api_client.processing",
            30,
            "[2023-12-12T13:00:00] This is a warning dataset message",
        ),
        (
            "cads_api_client.processing",
            20,
            "[2023-12-12T14:00:00] This is a success dataset message",
        ),
        (
            "cads_api_client.processing",
            10,
            (
                "POST http://localhost:8080/api/retrieve/v1/processes/"
                "reanalysis-era5-pressure-levels/execution "
                "{'variable': 'temperature', 'year': '2022'}"
            ),
        ),
        (
            "cads_api_client.processing",
            10,
            f"REPLY {json.dumps(JOB_SUCCESSFUL_JSON)}",
        ),
        (
            "cads_api_client.processing",
            20,
            "Request ID is 9bfc1362-2832-48e1-a235-359267420bb2",
        ),
        (
            "cads_api_client.processing",
            10,
            "GET http://localhost:8080/api/retrieve/v1/jobs/9bfc1362-2832-48e1-a235-359267420bb2",
        ),
        (
            "cads_api_client.processing",
            10,
            f"REPLY {json.dumps(JOB_SUCCESSFUL_JSON)}",
        ),
        ("cads_api_client.processing", 20, "This is a log"),
        ("cads_api_client.processing", 30, "This is a warning log"),
        ("cads_api_client.processing", 20, "status has been updated to successful"),
    ]



from __future__ import annotations

import datetime
from typing import Any

try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

import attrs
import requests

from . import config, processing


@attrs.define
class Collections(processing.ApiResponse):
    def collection_ids(self) -> list[str]:
        return [collection["id"] for collection in self.json["collections"]]

    def next(self) -> Self | None:
        return self.from_rel_href(rel="next")

    def prev(self) -> Self | None:
        return self.from_rel_href(rel="prev")


@attrs.define
class Collection(processing.ApiResponse):
    @property
    def temporal_interval(self) -> tuple[str, str]:
        begin, end = map(str, self.json["extent"]["temporal"]["interval"][0])
        return (begin, end)

    @property
    def begin_datetime(self) -> datetime.datetime:
        return datetime.datetime.fromisoformat(
            self.temporal_interval[0].replace("Z", "+00:00")
        )

    @property
    def end_datetime(self) -> datetime.datetime:
        return datetime.datetime.fromisoformat(
            self.temporal_interval[1].replace("Z", "+00:00")
        )

    @property
    def id(self) -> str:
        collection_id = self.json["id"]
        assert isinstance(collection_id, str)
        return collection_id

    def retrieve_process(
        self, headers: dict[str, str] | None = None
    ) -> processing.Process:
        url = self.get_link_href(rel="retrieve")
        kwargs = self.request_kwargs
        if headers is not None:
            kwargs["headers"] = headers
        return processing.Process.from_request("get", url, **kwargs)

    def submit(
        self,
        **request: Any,
    ) -> processing.Remote:
        retrieve_process = self.retrieve_process(headers=self.headers)
        status_info = retrieve_process.execute(inputs=request)
        return status_info.make_remote()

    def retrieve(
        self,
        target: str | None = None,
        **request: Any,
    ) -> str:
        remote = self.submit(**request)
        return remote.download(target)


@attrs.define(slots=False)
class Catalogue:
    url: str
    headers: dict[str, Any]
    session: requests.Session
    retry_options: dict[str, Any]
    request_options: dict[str, Any]
    download_options: dict[str, Any]
    sleep_max: int
    cleanup: bool
    force_exact_url: bool = False

    def __attrs_post_init__(self) -> None:
        if not self.force_exact_url:
            self.url += f"/{config.SUPPORTED_API_VERSION}"

    @property
    def request_kwargs(self) -> processing.RequestKwargs:
        return processing.RequestKwargs(
            headers=self.headers,
            session=self.session,
            retry_options=self.retry_options,
            request_options=self.request_options,
            download_options=self.download_options,
            sleep_max=self.sleep_max,
            cleanup=self.cleanup,
        )

    def collections(self, params: dict[str, Any] = {}) -> Collections:
        url = f"{self.url}/datasets"
        return Collections.from_request(
            "get", url, params=params, **self.request_kwargs
        )

    def collection(self, collection_id: str) -> Collection:
        url = f"{self.url}/collections/{collection_id}"
        return Collection.from_request("get", url, **self.request_kwargs)

    def licenses(self) -> dict[str, Any]:
        url = f"{self.url}/vocabularies/licences"
        return processing.ApiResponse.from_request(
            "get", url, **self.request_kwargs
        ).json
